(* CIS 341 *)
(* fun.ml  *)
(* Example code about interpretation and compilation of first class 
   (higher-order) functions. *)

(* A simple set implementation *)
module VarSet : sig
  type t
  val empty : t
  val union : t -> t -> t
  val singleton : string -> t
  val member : string -> t -> bool
  val remove : string -> t -> t
  val equals : t -> t -> bool
end = struct
  type t = string list
  let empty = []      
  let union s1 s2 = s1 @ s2
  let singleton x = [x]
  let rec member x s =
    match s with
    | [] -> false
    | y::ys -> x = y || member x ys
  let rec remove x s =
    match s with
    | [] -> []
    | y::ys -> if x = y then remove x ys else y::(remove x ys)

  let rec forall p s =
    List.fold_left (fun b x -> b && (p x)) true s
    
  let equals s1 s2 =
    (forall (fun x -> member x s1) s2) &&
    (forall (fun x -> member x s2) s1)

end


module Fun0 = struct 
  type var = string
 
  (* Abstract syntax of "lambda terms" *)
  type exp = 
    | Var of var                (* local variables *)
    | Fun of var * exp          (* functions:  fun x -> e *)
    | App of exp * exp          (* function application *)

  type value =
    | FunV of var * exp

  (* values are just a subset of expressions *)
  let exp_of_value v = 
    match v with
      | FunV (arg, body) -> Fun(arg, body)



  
  (* Example untyped lambda calculus terms *)

  let ex1 = Fun("x", Var "x")
  let ex2 = App(ex1, ex1)

  
  (* The omega term -- demonstrates an infinite loop written in
     untyped lambda calculus. *)
  let omega =
    App(Fun ("x", App(Var "x", Var"x")),
	Fun ("x", App(Var "x", Var"x")))



  
  (* The "Y Combinator" -- not a startup incubator!  It calculates
     fixedpoints of a function f *)
  let y_combinator =
    Fun ("f", App(Fun("x", App(Var "f", App(Var "x", Var "x"))),
		  Fun("x", App(Var "f", App(Var "x", Var "x")))))


  let true_l = Fun("x", Fun("y", Var "x"))
  let false_l = Fun("x", Fun("y", Var "y"))


  
  (* Church encoding of the natural numbers *)
  let zero = Fun("f", Fun("y", Var "y"))
  let one  = Fun("f", Fun("y", App(Var "f", Var "y")))
  let two  = Fun("f", Fun("y", App(Var "f", App(Var "f", Var "y"))))
  let mk_n n =
    let rec apps n = if n = 0 then Var "y" else App(Var "f", apps (n-1)) in
    Fun("f", Fun("y", apps n))

  
  (* An arithmetic operation *)
  let mult =  Fun("x1", Fun("x2", Fun("f", Fun("y", App(App(Var "x1", App(Var "x2", Var "f")), Var "y")))))
  let test_mult = App(App(mult, two), two)


  let pair x y = Fun("pair", App(App(Var "pair", x), y))
  let fst e = App(e, Fun("x", Fun("y", Var "x")))
  let snd e = App(e, Fun("x", Fun("y", Var "y")))      

end


(* A substitution-based evaluator *)
module Eval0 = struct
  open Fun0

  (* Substitute the (closed) value v for x in e *)
  let rec subst v x e =
    match e with
      | Var y -> if y = x then exp_of_value v else Var y
      | Fun (arg, body) -> 
	  if (arg = x) then Fun (arg, body) 
	  else Fun(arg, subst v x body)
      | App (e1, e2) -> App(subst v x e1, subst v x e2)


  (* A substituion-based interpreter - assumes e is closed 

     Compare this code to the large-step operational semantics rules:

       -----------
        v  ==>  v 

 
       e1 ==> (fun x -> e3)    e2 ==> v     e3{v/x} ==> w
      ----------------------------------------------------     
       e1 e2 ==> w
  *)
  let rec eval e =
    match e with
      | Var x -> failwith "free variable: impossible if e is closed"
      | Fun (arg, body) -> FunV (arg, body)
      | App (e1, e2) -> 
	  (match(eval e1, eval e2) with
	     | (FunV (x, body), v) -> eval (subst v x body))

end





module Fun = struct 
  type var = string
 
  (* Abstract syntax of "lambda terms" *)
  type exp = 
      Int of int              
    | Add of exp * exp
    | Var of var                (* local variables *)
    | Fun of var * exp          (* functions:  fun x -> e *)
    | App of exp * exp          (* function application *)

  type value =
    | IntV of int
    | FunV of var * exp

  (* values are just a subset of expressions *)
  let exp_of_value v = 
    match v with
      | IntV i -> Int i
      | FunV (arg, body) -> Fun(arg, body)

  let rec free_vars (e:exp) : VarSet.t =
    match e with
    | Int _ -> VarSet.empty
    | Var x -> VarSet.singleton x
    | App(e1,e2)
    | Add(e1, e2) -> VarSet.union (free_vars e1) (free_vars e2)
    | Fun(x, e)   -> VarSet.remove x (free_vars e)

  
  (* Example term: *)
  (* ( ( (fun x -> fun y -> x + y) 3 ) 4 ) *)
  let ex1 = App (App (Fun ("x", Fun ("y", Add (Var "x", Var "y"))), 
                     Int 3), 
                Int 4)


  (* fun x -> fun y -> fun z -> x + y + z *)
  let ex2 =
    Fun ("x", Fun("y", Fun("z", Add(Var "x", Add(Var "y", Var "z")))))
  
  let ex3 =
    App (ex2, Int 2)

  let ex4 =
    App (ex3, Int 3)
  
  (* The following example illustrates the difference between dynamically
     scoped and statically scopes variable binding:
     - under static scoping (i.e. substitution semantics), it returns
       the value 3
     - under dynamic scoping, it returns the value 5

     ( (fun f -> ((fun x -> f 1) 4))
       ( (fun x -> fun y -> x + y) 2) )
  *)
  let dyn = 
  App(Fun ("f", App( Fun ("x", App(Var "f", Int 1)), Int 4)),
       App(Fun ("x", (Fun ("y", Add (Var "x", Var "y")))), Int 2))




  

  (* The omega term -- demonstrates an infinite loop written in
     untyped lambda calculus. *)
  let omega =
    App(Fun ("x", App(Var "x", Var"x")),
	Fun ("x", App(Var "x", Var"x")))

  (* The "Y Combinator" -- not a startup incubator!  It calculates
     fixedpoints of a function f *)
  let y_combinator =
    Fun ("f", App(Fun("x", App(Var "f", App(Var "x", Var "x"))),
		  Fun("x", App(Var "f", App(Var "x", Var "x")))))

  (* Church encoding of the natural numbers *)
  let zero = Fun("f", Fun("y", Var "y"))
  let one  = Fun("f", Fun("y", App(Var "f", Var "y")))
  let two  = Fun("f", Fun("y", App(Var "f", App(Var "f", Var "y"))))
  let mk_n n =
    let rec apps n = if n = 0 then Var "y" else App(Var "f", apps (n-1)) in
    Fun("f", Fun("y", apps n))

  let pair x y = Fun("pair", App(App(Var "pair", x), y))
  let fst e = App(e, Fun("x", Fun("y", Var "x")))
  let snd e = App(e, Fun("x", Fun("y", Var "y")))      
  
  (* An arithmetic operation *)
  let mult =  Fun("x1", Fun("x2", Fun("f", Fun("y", App(App(Var "x1", App(Var "x2", Var "f")), Var "y")))))
  let test_mult = App(App(mult, two), two)

  
  let church_to_int c = App(App(c, Fun("z", Add(Var "z", Int 1))), Int 0)  
end


(* A substitution-based evaluator *)
module Eval1 = struct
  open Fun

  (* Substitute the (closed) value v for x in e *)
  let rec subst v x e =
    match e with
      | Int i -> Int i
      | Add (e1, e2) -> Add(subst v x e1, subst v x e2)
      | Var y -> if y = x then exp_of_value v else Var y
      | Fun (arg, body) -> 
	  if (arg = x) then Fun (arg, body) 
	  else Fun(arg, subst v x body)
      | App (e1, e2) -> App(subst v x e1, subst v x e2)

  (* A substituion-based interpreter *)
  let rec eval e =
    match e with
      | Int i -> IntV i
      | Add (e1, e2) -> 
	  (match (eval e1, eval e2) with
	    | (IntV i1, IntV i2) -> IntV (i1 + i2)
	    | _ -> failwith "tried to add non-integers")
      | Var x -> failwith "reached a free variable"
      | Fun (arg, body) -> FunV (arg, body)
      | App (e1, e2) -> 
	  (match(eval e1, eval e2) with
	     | (FunV (x, body), v) -> eval(subst v x body)
	     | _ -> failwith "tried to apply non-function")

end



(* A *broken* attempt to use an environment-based evaluator *)
(* This is broken in the sense that it disagrees with the 
   substitution semantics presented in the Eval1 module. *)
module Eval2 = struct
  open Fun

  type environment = (var * value) list
  let lookup x (env:environment) = List.assoc x env

  (* The following evaluator gives *dynamic* scoping, not
     static (or 'lexical') scoping! *)
  let rec eval env e =
    match e with 
      | Int i -> IntV i
      | Add (e1, e2) ->
	  (match (eval env e1, eval env e2) with
	     | (IntV i1, IntV i2) -> IntV (i1 + i2)
	     | _ -> failwith "tried to add non-integers")
      | Var x -> lookup x env
      | Fun (arg, body) -> FunV(arg, body)
      | App (e1, e2) ->
	  (match(eval env e1, eval env e2) with
	     | (FunV (x, body), v) -> eval ((x,v)::env) body 
	     | _ -> failwith "tried to apply non-function")

end





(* A "closure"-based interpreter that fixes the problem with Eval2 *)
(* A closure is a pair consisting of a datastructure that represents 
   the evaluation context and a function *)
module Eval3 = struct
  open Fun

  (* Change the definition of value... *)
  type value =
    | IntV of int
    | Closure of environment * var * exp
  and environment = (var * value) list
      
  let lookup x (env:environment) = List.assoc x env

  (* The following evaluator fixes the problems with Eval2 *)
  (* It yields statically-scoped  variables *)
  let rec eval env e  =
    match e with 
      | Int i -> IntV i
      | Add (e1, e2) ->
	  (match (eval env e1, eval env e2) with
	     | (IntV i1, IntV i2) -> IntV (i1 + i2)
	     | _ -> failwith "tried to add non-integers")
      | Var x -> lookup x env

      (* NOTE: the closure captures the environment in force at 
         the time that the function is created *)
      | Fun (arg, body) -> Closure(env, arg, body)

      | App (e1, e2) ->
	  (match(eval env e1, eval env e2) with
	     | (Closure(cenv, x, body), v) -> eval ((x,v)::cenv) body 
	     | _ -> failwith "tried to apply non-function")

end













